# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluation metrics."""
import dataclasses
import logging
import math
from typing import Optional, List
from lsmmdma.mmdma_functions import compute_sqpairwise_distances
import numpy as np
from pykeops.torch import LazyTensor
from sklearn.neighbors import KNeighborsClassifier
import torch


@dataclasses.dataclass()
class OutputSupervised():
  foscttm: float
  top1: float
  top5: float
  no: float
  lta: float


# Evaluation.
@dataclasses.dataclass()
class Evaluation():
  """Computes evaluation metrics, knowing the ground truth alignment.

  Five metrics are being computed:
  - FOSCTTM (Fraction Of Samples Closer Than the True Match): it gives the
  average number of samples from one view that are closer to the true match
  in the other view, in the learned low dimensional space. As this metrics is
  not symmetrical if we exchange the first and second views, the results are
  averaged calculating the FOSCTTM in both directions. Two implementations are
  available, `_foscttm` and `_foscttm_keops`, the latter scaling to several
  hundreds of samples.
  - topK (top1 and top5): it computes the fraction of samples from one view
  correctly assigned among top K% nearest neighbours of the true match from the
  other view. As this metrics is not symmetrical if we exchange the first and
  second views, the results are averaged calculating top1 in both directions.
  K is set to be 1 and 5 by default.
  - Neighbourhood Overlap: it computes the fraction of samples from one view
  correctly assigned among the top K nearest neighbours of the true match from
  the other view, for all K. The output is a vector of dimension K.
  - Label Transfer Accuracy: assigns the label to samples from the second_view
  (resp. first_view) based on the 5-nearest neighbours from the first_view
  (resp. seconv_view). Calculates accuracy of the predicted labels.
  """
  ground_truth_alignment: Optional[np.ndarray] = None
  cell_labels: Optional[List[np.ndarray]] = None
  n_neighbours: int = 5
  device: torch.device = torch.device('cuda')
  short: bool = False

  def compute_all_evaluation(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor
      ) -> OutputSupervised:
    """Computes evaluation measures, knowing the ground truth alignment."""
    if self.ground_truth_alignment is not None:
      second_view_aligned = second_view[self.ground_truth_alignment]

    n = first_view.shape[0]
    m = second_view.shape[0]

    foscttm = -1
    top1 = -1
    top5 = -1
    neigh_overlap = -1
    lta = -1

    if self.ground_truth_alignment is not None:
      try:
        if n == m and n < 5000:
          foscttm = self._foscttm(first_view, second_view_aligned)
        elif n == m:
          foscttm = self._foscttm_keops(first_view, second_view_aligned)
        else:
          logging.warning(f'FOSCTTM was not computed because {n} != {m}.')
      except:
        logging.warning(
            'FOSCTTM was not computed and most likely led to an OOM issue.')

      if not self.short:
        try:
          top1 = self._topx_keops(
              first_view, second_view_aligned, topk=1, percentage=True)
        except:
          logging.warning(
              'Top1 can not be computed with a number of samples <100.')

        try:
          top5 = self._topx_keops(
              first_view, second_view_aligned, topk=5, percentage=True)
        except:
          logging.warning(
              'Top5 can not be computed with a number of samples <20.')

        try:
          if n == m:
            if n * m < 1e7:
              neigh_overlap = self._neighbour_overlap(
                  first_view, second_view_aligned)
            else:
              logging.warning("""Switching to batch version""")
              neigh_overlap = self._neighbour_overlap_batch(
                  first_view, second_view_aligned)
        except:
          logging.warning(f"""Neighbourhood overlap was not computed, either
                          because {n} != {m} or because it might have led to an
                          OOM issue""")
      else:
        logging.info("""TopK and NO were not computed.""")

    if self.cell_labels is not None:
      try:
        lta1 = self._label_transfer_accuracy(
            first_view, second_view, self.cell_labels, k=self.n_neighbours)
        lta2 = self._label_transfer_accuracy(
            second_view, first_view, self.cell_labels[::-1],
            k=self.n_neighbours)
        lta = (lta1 + lta2) / 2.0
      except:
        logging.warning("""LTA was not computed because it might have led to
                        an OOM issue""")

    return OutputSupervised(foscttm, top1, top5, neigh_overlap, lta)

  def _foscttm(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor
      ) -> float:
    """Computes the fraction of samples closer to the true match based on
    squared euclidean distances between samples."""
    n = first_view.shape[0]
    # Assumes the views are aligned.
    distances = compute_sqpairwise_distances(first_view, second_view)
    fraction = (
        torch.sum(distances < torch.diag(distances))
        + torch.sum(torch.t(distances) < torch.diag(distances))
        ) / (2 * n * (n - 1))
    return fraction.item()

  def _foscttm_keops(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor,
    ) -> float:
    """Computes the fraction of samples closer to the true match based on
    squared euclidean distances between samples."""
    n, d = first_view.shape

    first_view_i = LazyTensor(first_view.view(n, 1, d))
    second_view_j = LazyTensor(second_view.view(1, n, d))

    distance_ij = ((first_view_i - second_view_j)**2).sum(dim=2)

    diagonal = ((first_view - second_view)**2).sum(axis=1)
    diagonal1 = LazyTensor(diagonal.view(-1, 1, 1))
    diagonal2 = LazyTensor(diagonal.view(1, -1, 1))

    cttm1 = (diagonal1 - distance_ij).sign().relu()
    cttm2 = (diagonal2 - distance_ij).sign().relu()
    cttm1 = cttm1.sum(1)
    cttm2 = cttm2.sum(1)
    return (cttm1.sum() + cttm2.sum()) / (n * (n - 1) * 2)

  def _topx_keops(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor,
      topk: int = 1,
      percentage: bool = True
      ) -> float:
    """Computes fraction of samples correctly assigned among topk (%) nearest
    neighbours."""
    def get_count_knn(distance: LazyTensor, n_sample, dim: int = 0):
      # Grid <-> Samples, (M**2, K) integer tensor.
      if percentage:
        indknn = distance.argKmin(int(topk * n_sample / 100), dim=dim)
      else:
        indknn = distance.argKmin(topk, dim=dim)
      frac = indknn - torch.arange(n_sample).reshape(-1, 1).to(self.device)
      return torch.count_nonzero(frac == 0).item()

    n = first_view.shape[0]
    m = second_view.shape[0]
    first_view_i = LazyTensor(first_view[:, None, :])  # (M**2, 1, 2)
    second_view_j = LazyTensor(second_view[None, :, :])  # (1, N, 2)
    # (M**2, N) symbolic matrix of squared distances.
    distance_ij = ((first_view_i - second_view_j) ** 2).sum(-1)
    count_nn0 = get_count_knn(distance_ij, n, dim=0)
    count_nn1 = get_count_knn(distance_ij, m, dim=1)
    return (count_nn0 / m + count_nn1 / n) / 2.0

  def _neighbour_overlap(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor,
      ) -> float:
    """Computes the neighbourhood overlap metric.

    The Neihgbourhood Overlap is the fraction of samples that have among
    their k-neighbours the true match. This function assumes that the two
    datasets are aligned (sample1 in first_view corresponds to sample1 in
    second_view) and have the same number of cells.
    """
    n = first_view.shape[0]
    distances = compute_sqpairwise_distances(first_view, second_view)
    ranking = torch.diagonal(torch.argsort(distances, dim=1))
    cumul_rank = [torch.sum(ranking <= rank).item() for rank in range(n)]
    cumul_frac = np.array(cumul_rank) / n
    return cumul_frac

  def _neighbour_overlap_batch(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor,
      batch_size: int = 1024,
      ) -> float:
    """Computes the neighbourhood overlap metrics by batches of the first_view.

    The Neihgbourhood Overlap is the fraction of samples that have among
    their k-neighbours the true match. This function assumes that the two
    datasets are aligned (sample1 in first_view corresponds to sample1 in
    second_view) and have the same number of cells.
    """
    n = first_view.shape[0]
    n_batch = math.ceil(n / batch_size)
    cumul_frac = np.zeros(n)
    for i in range(n_batch):
      distances = compute_sqpairwise_distances(
          first_view[i * batch_size: (i+1) * batch_size], second_view)
      ranking = torch.diagonal(torch.argsort(distances, dim=1))
      cumul_rank = [torch.sum(ranking <= rank).item() for rank in range(n)]
      cumul_frac += np.array(cumul_rank)
    return cumul_frac / n

  def _label_transfer_accuracy(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor,
      cell_labels: List[np.ndarray],
      k: int = 5):
    """Computes the Label Transfer Accuracy metrics."""
    first_view = first_view.detach().cpu()
    second_view = second_view.detach().cpu()
    knn = KNeighborsClassifier(n_neighbors=k)
    knn.fit(first_view, cell_labels[0])
    predictions = knn.predict(second_view)
    count = 0
    for label1, label2 in zip(predictions, cell_labels[1]):
      if label1 == label2:
        count += 1
    return count / second_view.shape[0]




# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates data with varying number of samples and features."""
import os
import random
from typing import Tuple, Union

from lsmmdma.data.checkpointer import save_generated_data
import numpy as np
import scanpy
import torch
from tensorflow.io import gfile


def generate_data(
    path: str,
    n_sample: int,
    p_feature: int,
    random_seed: int = 4,
    simulation: str = 'branch',
    ) -> Union[Tuple[np.ndarray], Tuple[torch.Tensor]]:
  """Generates two views following the `simulation` shape in their latent space.

  The simulation process follows the one given in the original MMD-MA paper (Liu
  et al. https://pubmed.ncbi.nlm.nih.gov/34632462/). A manifold (of shape
  'branch' or 'triangle') is generated in two dimensions. The resulting set of
  points is standardised. The points are then mapped to `p_feature` dimensions
  using a (2 x p_feature) mapping, sampled from a standard Gaussian distribution,
  resulting in a (n_sample * p_feature) matrix. Gaussian noise is then added to
  each element of the matrix.

  In these simulations, we assume that the numbers of samples and features are
  are the same for both domains.

  Arguments:
    path: path to output directory where to save the generated data.
    n_sample: number of samples.
    p_feature: number of features.
    random_seed: seed for random (random_seed) and numpy (random_seed + 1)
    simulation: str, latent structure for the input data. The latent structure
      can either be a 'branch' (two elongated Gaussian point clouds
      perpendicular to each other) or a 'triangle' (three elongated Gaussian
      point clouds forming a triangle).

  Returns:
    tuple, first view, second view and corresponding indices from the first
      view.
  """
  latent_dim = 2
  # TODO(lpapaxanthos): allow for different number of features.
  p_features1 = p_feature
  p_features2 = p_feature
  noise = 0.05

  if simulation == 'triangle':
    means = [(0, 1), (3.5, 4.33), (-3.5, 4.33)]
    covs = [[[8, 0], [0, 0.05]],
            [[1.7, -1.5], [-1.5, 1.7]],
            [[1.7, 1.5], [1.5, 1.7]]]
    n1 = int(n_sample / 4)
    shapes = [2 * n1, n1, n1 - 3 * n1]
    rot = None
  elif simulation == 'branch':
    means = [[0, 0], [0, 8]]
    covs = [[[20, 0], [0, 1]], [[1, 0], [0, 10]]]
    n1 = int(n_sample / 4 * 3)
    shapes = [n1, n_sample - n1]
    rot = np.array([[np.cos(np.pi/4.), -np.sin(np.pi/4.)],
                    [np.sin(np.pi/4.), np.cos(np.pi/4.)]])
  else:
    raise ValueError('Value for simulation is not supported')

  length = len(means)

  random.seed(random_seed)
  np.random.seed(random_seed + 1)
  axis = [0] * length
  for i in range(length):
    axis[i] = np.random.multivariate_normal(
        mean=np.array(means[i]), cov=np.array(covs[i]), size=shapes[i])

  latent_array = np.concatenate((list(axis[i] for i in range(length))))
  if rot is not None:
    latent_array = np.dot(latent_array, rot)
  latent_array = (latent_array - latent_array.mean(0)) / latent_array.std(0)

  noise_first_view = np.random.normal(
      size=(latent_dim, p_features1))
  noise_second_view = np.random.normal(
      size=(latent_dim, p_features2))
  first_view = (torch.FloatTensor(
      np.dot(latent_array, noise_first_view) + np.random.normal(
          size=(n_sample, p_features1)) * noise))
  second_view = (torch.FloatTensor(
      np.dot(latent_array, noise_second_view) + np.random.normal(
          size=(n_sample, p_features2)) * noise))

  rd_vec = np.random.choice(n_sample, n_sample, replace=False)
  first_view = first_view[rd_vec]

  if path:
    save_generated_data(path, first_view, second_view, rd_vec)
  return first_view, second_view, rd_vec


def load(input_dir: str, filename: str) -> np.ndarray:
  """Loads data."""
  path_file = os.path.join(input_dir, filename)
  file_ext = os.path.splitext(path_file)[-1]
  with gfile.GFile(path_file, 'rb') as my_file:
    if file_ext == '.h5ad':
      data = scanpy.read_h5ad(my_file)
      data = data.X.todense()
    elif file_ext == '.npy':
      data = np.load(my_file)
    elif file_ext == '.tsv':
      data = np.loadtxt(my_file, delimiter='\t')
    elif file_ext == '.csv':
      data = np.loadtxt(my_file, delimiter=',')
    else:
      try:
        data = np.loadtxt(my_file)
      except:
        raise ValueError('Input data is not in a supported format.')
  return data



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of a large-scale version of MMD-MA with PyTorch.

MMD-MA has been developped by Liu et al. 2019, Jointly Embedding Multiple
Single-Cell Omics Measurements (https://pubmed.ncbi.nlm.nih.gov/34632462/).
"""

import random
from absl import logging

from collections import defaultdict as ddict
import dataclasses
from lsmmdma.initializers import initialize
from lsmmdma.metrics import Evaluation
import lsmmdma.mmdma_functions as mmdma_fn
from math import ceil
import numpy as np
from tenacity import retry, stop_after_attempt
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
from torch.utils import tensorboard
from typing import Tuple, Any, Dict, DefaultDict, List, Union


@dataclasses.dataclass(unsafe_hash=True)
class ModelGetterConfig:
  """Model attributes."""
  seed: int
  learning_rate: float = 1e-3
  low_dim: int = 5
  n_record: int = 10
  n_eval: int = 10
  sigmas: float = 1.
  lambda1: float = 1.
  lambda2: float = 1.
  n_iter: int = 1000
  mode: str = 'dual'
  keops: bool = False
  pca: bool = False
  n_seed: int = 1
  init: str = 'uniform'
  use_unbiased_mmd: bool = True


class Model(nn.Module):
  """Model for MMD-MA, defines learned parameters and computes the loss.

  In MMD-MA, we project two sets of points, from two different spaces endowed
  with a positive definite kernel, to a shared Euclidean space of dimension
  low_dim. The mappings from high to low dimensional space is
  obtained using functions belonging to the respective RKHS. To obtain the
  mappings, we minimise a loss function that is composed of three terms:
  - an MMD term between the low dimensional representations of the two views,
  which encourages them to have the same distribution.
  - two non-collapsing penalty terms (corresponding to the pen_dual or
  pen_primal functions), one for each view. These terms ensure that
  the low dimensional representations are mutually orthogonal, preventing
  collapsing.
  - two distortion penalties (corresponding to the dis_dual or
  dis_primal functions), one for each view. These terms encourage the
  low dimensional representation to obtain the same pairwise structure as
  the original views.

  MMD-MA can be formulated using either the primal (when we use the linear
  kernel in the input spaces) or the dual problem. Each has
  advantages or disadvantages depending on the input data. Let D1 and D2 be two
  sets of points, of respective size n_sample1*p_feature1 and
  n_sample2*p_feature2, where n_sample1 (resp. n_sample2) corresponds to
  the number of samples of the first set of points D1 (resp.
  the second set of points D2) and p_feature1 (resp. p_feature2)
  corresponds to the number of features of the first set of points D1
  (resp. the second one D2). In each view, when the
  number of features is larger than the number of samples
  p_featureX >> n_sampleX, then the dual formulation is beneficial in terms
  of runtime and memory, while if n_sampleX >> p_sampleX, the primal
  formulation is favorable.

  The original implementation of MMD-MA (see Liu et al.) uses the dual
  formulation. The resulting loss can be written as follows:
  min_{W1, W2} Loss(W1, W2)
        = MMD(K1W1, K2W1)
        + lambda1(||W1.TK1.TW1 - I|| + ||W2.TK2.TW2 - I||)
        + lambda2(||K1W1W1.TK1.T - K1|| + ||K2W2W2.TK2.T - K2||)
  where W1 and W2 are the set of weights that we wish to learn, K1 and K2 are
  kernels of the input views D1 and D2, and I are identity matrices. The dual
  form has the advantage of only involving the kernels K1 and K2 and not the
  original sets of points D1 ans D2. Therefore, non linear mappings can also
  be used.

  If the number of samples is much larger than the number of features in both
  views, then the primal formulation can be favorable:
  min_{W1, W2} Loss(W1, W2)
        = MMD(D1W1, D2W1)
        + lambda1(||W1.TW1 - I|| + ||W2.TW2 - I||)
        + lambda2(||D1W1W1.TD1.T - D1D1.T|| + ||D2W2W2.TD2.T - D2D2.T||)
  where W1 and W2 are the model parameters.

  Attributes:
    cfg_model: ModelGetterConfig, sets parameters for the MMDMA algorithm.
    n_sample1: int, number of samples in the first view.
    n_sample2: int, number of samples in the second view.
    p_feature1: int, number of features in the first view.
    p_feature2: int, number of features in the second view.
    keops: bool, whether to use keops when calculating MMD.
    device: torch.device, whether a cpu or gpu is used.
  """

  def __init__(
      self,
      cfg_model: ModelGetterConfig,
      n_sample1: int,
      n_sample2: int,
      p_feature1: int,
      p_feature2: int,
      keops: bool,
      device: torch.device,
      ):
    super(Model, self).__init__()
    self.n_sample1 = n_sample1
    self.n_sample2 = n_sample2
    self.p_feature1 = p_feature1
    self.p_feature2 = p_feature2
    self.cfg_model = cfg_model
    self.keops = keops
    self.device = device

    if self.cfg_model.mode == 'primal':
      param1_dim = self.p_feature1
      param2_dim = self.p_feature2
    elif self.cfg_model.mode == 'dual':
      param1_dim = self.n_sample1
      param2_dim = self.n_sample2
    else:
      raise NotImplementedError

    init_args = cfg_model.init.split(',')
    init_args[1:] = [float(i) for i in init_args[1:]]
    initializer = initialize(init_args[0])
    empty_tensor1 = torch.empty(param1_dim, cfg_model.low_dim)
    empty_tensor2 = torch.empty(param2_dim, cfg_model.low_dim)
    init1 = initializer(empty_tensor1, *init_args[1:])
    init2 = initializer(empty_tensor2, *init_args[1:])
    self.weights_first_view = Parameter(init1)
    self.weights_second_view = Parameter(init2)

  def forward(self,
              first_view: torch.Tensor,
              second_view: torch.Tensor
              ) -> Tuple[float, Tuple[float, ...]]:
    """Forward pass.

    Arguments:
      first_view: torch.Tensor, first view. The dimension of this tensor is
        (sample x feature) if mode == 'primal' and (sample x sample)
        if mode == 'dual'.
      second_view: torch.Tensor, second view. The dimension of this tensor is
        (sample x feature) if mode == 'primal' and (sample x sample)
        if mode == 'dual'.

    Returns:
      loss: float, MMDMA loss
      loss_components: 5-tuple, individual components of the MMDMA loss
    """
    loss, loss_components = loss_mmdma(
        [self.weights_first_view, self.weights_second_view],
        first_view, second_view, self.n_sample1, self.n_sample2,
        self.cfg_model, self.keops, self.device)
    return loss, loss_components


def loss_mmdma(
    params: torch.Tensor,
    first_view: torch.Tensor,
    second_view: torch.Tensor,
    n_first_view: int,
    n_second_view: int,
    cfg_model: ModelGetterConfig,
    keops: bool,
    device: torch.device
    ) -> Tuple[float, Tuple[float, ...]]:
  """Loss of the MMDMA algorithm.

  The loss is composed of an MMD term to which two penalty terms are added. The
  first penalty type (`penalty_xv`) penalises collapsing embedding
  spaces and the second penalty type (`distortion_xv`) penalises
  distortion between the original space and the embedding space.

  Arguments:
    params: torch.Tensor, model parameters.
    first_view: torch.Tensor, first set of points. The dimension of this
      tensor is (sample x feature) if mode == 'primal' and (sample x sample)
      if mode == 'dual'.
    second_view: torch.Tensor, second set of points. The dimension of this
      tensor is (sample x feature) if mode == 'primal' and (sample x sample)
      if mode == 'dual'.
    n_first_view: int, number of samples in the first view.
    n_second_view: int, number of samples in the second view.
    cfg_model: ModelGetterConfig, sets parameters for the MMDMA algorithm.
    keops: bool, whether to use keops when calculating MMD.
    device: torch.device, whether a cpu or gpu ('cuda') is used.

  Returns:
    tuple containing the loss and the individual components of the loss
  """
  embedding_fv = torch.matmul(first_view, params[0])
  embedding_sv = torch.matmul(second_view, params[1])
  mmd = mmdma_fn.squared_mmd(
      embedding_fv, embedding_sv, cfg_model.sigmas, keops,
      cfg_model.use_unbiased_mmd, device)
  if cfg_model.mode == 'primal':
    penalty_fv = mmdma_fn.pen_primal(params[0], device)
    penalty_sv = mmdma_fn.pen_primal(params[1], device)
    distortion_fv = mmdma_fn.dis_primal(
        first_view, params[0], n_first_view)
    distortion_sv = mmdma_fn.dis_primal(
        second_view, params[1], n_second_view)
  elif cfg_model.mode == 'dual':
    penalty_fv = mmdma_fn.pen_dual(embedding_fv, params[0], device)
    penalty_sv = mmdma_fn.pen_dual(embedding_sv, params[1], device)
    distortion_fv = mmdma_fn.dis_dual(embedding_fv, first_view)
    distortion_sv = mmdma_fn.dis_dual(embedding_sv, second_view)
  else:
    raise ValueError('cfg_model.mode must be "dual" or "primal."')
  return (mmd + cfg_model.lambda1 * (penalty_fv + penalty_sv)
          + cfg_model.lambda2 * (distortion_fv + distortion_sv),
          (mmd.item(), penalty_fv.item(), penalty_sv.item(),
           distortion_fv.item(), distortion_sv.item()))


def get_kernel(data: torch.Tensor, kernel_type='linear') -> torch.Tensor:
  """Gives linear kernel needed in the dual mode.

  Arguments:
    data: torch.Tensor, one of the two views (sample x feature).
    kernel_type: str, 'linear' by default but other kernels can be implemented.

  Returns:
    kernel: torch.Tensor, linear kernel of one of the two views
      (sample x sample).
  """
  if kernel_type == 'linear':
    kernel = torch.matmul(data, data.T)
  else:
    raise NotImplementedError(f'{kernel_type} kernel is not supported yet.')
  return kernel


def save_loss_value(
    loss_tmp: float,
    loss_components_tmp: List[float],
    evaluation: Dict[str, float],
    ) -> Dict[str, float]:
  """Records loss related metrics.

  Arguments:
    loss_tmp: float, loss value.
    loss_components_tmp: tuple, components of the loss.
    evaluation: dictionary, records the output values.

  Returns:
    evaluation: dictionary that records the output values.
  """
  evaluation['loss'].append(loss_tmp.item())
  for i, val in enumerate(
      ['mmd', 'pen_fv', 'pen_sv', 'dis_fv', 'dis_sv']):
    evaluation[val].append(loss_components_tmp[i])
  return evaluation


def save_evaluation(
    evaluation_tmp: Dict[str, float],
    evaluation: Dict[str, float]
    ) -> Dict[str, float]:
  """Appends evaluation values to dictionary.

  Arguments:
    evaluation_tmp: dictionary, current evaluation values.
    evaluation: dictionary, records the output values.

  Returns:
    evaluation: dictionary that records the output values.
  """
  for key, val in evaluation_tmp.items():
    evaluation[key].append(val)
  return evaluation


def _evaluate(
    i: int,
    first_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
    second_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
    model: torch.nn.Module,
    eval_fn: Evaluation,
    loss: List[float],
    loss_components: List[List[float]],
    evaluation_loss: DefaultDict[str, List[float]],
    evaluation_matching: DefaultDict[str, List[float]],
    embeddings_results: Tuple[np.ndarray],
    pca_results: Tuple[np.ndarray],
    cfg_model: ModelGetterConfig,
    device: torch.device,
    summary_writer: tensorboard.SummaryWriter,
    workdir: str
    ) -> Tuple[DefaultDict[str, List[float]],
               DefaultDict[str, List[float]], Tuple[np.ndarray]]:
  """Records loss and evaluation metrics, runs PCA on embeddings.

  Arguments:
    i: number of epoch.
    first_view: first point cloud (sample x feature) if primal and first kernel
      (sample x sample) if dual.
    second_view: second point cloud (sample x feature) if primal and second
      kernel (sample x sample) if dual.
    model: mmdma model.
    eval_fn: metrics to evaluate the model.
    loss: loss.
    loss_components: loss components.
    evaluation_loss: records the loss and loss components
      during training.
    evaluation_matching: records the metrics during training.
    embeddings_results: records embeddings during training.
    pca_results: records results of PCA on embeddings during training.
    cfg_model: contains the parameters of the model and algorithm.
    device: either torch.device('cuda') or 'cpu'.
    summary_writer: summary writer for tensorboard.
    workdir: directory where the files are saved.

  Returns:
    evaluation_loss: records the loss and loss components
      during training.
    evaluation_matching: records the metrics during training.
    pca_results: records results of PCA on embeddings during training.
  """
  @retry(stop=stop_after_attempt(3))
  def init_summary():
    if (i == 0
        and (cfg_model.n_record != 0 or cfg_model.n_eval != 0) and workdir):
      summary_writer = tensorboard.SummaryWriter(workdir)
      return summary_writer
    else:
      return None

  if i == 0:
    summary_writer = init_summary()

  weights1, weights2 = list(model.parameters())
  embeddings_fv = torch.matmul(first_view, weights1)
  embeddings_sv = torch.matmul(second_view, weights2)

  # Records loss.
  if cfg_model.n_record != 0 and i % cfg_model.n_record == 0:
    evaluation_loss = save_loss_value(
        loss, loss_components, evaluation_loss)
    if workdir:
      for key, val in evaluation_loss.items():
        summary_writer.add_scalar(f'train/{key:s}', val[-1], i)
      logging.info('Flushing TensorBoard writer.')
      summary_writer.flush()

  # Saves evaluation metrics.
  if cfg_model.n_eval != 0 and i % cfg_model.n_eval == 0:
    eval_out = eval_fn.compute_all_evaluation(embeddings_fv, embeddings_sv)
    eval_out_dict = dataclasses.asdict(eval_out)
    logging.info('Train evaluation: %s.', eval_out)
    if workdir:
      for key, val in eval_out_dict.items():
        summary_writer.add_scalar(f'train/{key:s}', val, i)
        logging.info('Flushing TensorBoard writer.')
    evaluation_matching = save_evaluation(
        eval_out_dict, evaluation_matching)

  # Saves low dimensional representation
  if ((cfg_model.n_eval != 0 and i % cfg_model.n_eval == 0)
      or (cfg_model.n_record != 0 and i % cfg_model.n_record == 0)):
    embeddings_results.append(
        [embeddings_fv.detach().cpu().numpy(),
         embeddings_sv.detach().cpu().numpy()])

  # Saves 2-dimensional representation obtained with PCA.
  if cfg_model.pca != 0 and i % cfg_model.pca == 0:
    pca_rep_fv, pca_rep_sv = mmdma_fn.pca(
        embeddings_fv.detach().cpu().numpy(),
        embeddings_sv.detach().cpu().numpy())
    pca_results.append([pca_rep_fv, pca_rep_sv])

  if (i == cfg_model.n_iter - 1
      and (cfg_model.n_record != 0 or cfg_model.n_eval != 0) and workdir):
    summary_writer.close()
  return (evaluation_loss,
          evaluation_matching,
          embeddings_results,
          pca_results,
          summary_writer)


def train_and_evaluate(
    cfg_model: ModelGetterConfig,
    first_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
    second_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
    eval_fn: Evaluation,
    workdir: str,
    device: torch.device,
    ) -> Any:
  """Computes the training loop and evaluation steps.

  Arguments:
    cfg_model: ModelGetterConfig, sets parameters for the MMDMA algorithm.
    first_view: torch.Tensor, first view (sample1 x feature1) if primal and
      (sample1 x sample1) if dual.
    second_view: torch.Tensor, second view (sample2 x feature2) if primal and
      (sample2 x sample2) if dual.
    eval_fn: Evaluation dataclass, applies functions for evaluation.
    workdir: str, path to output directory.
    device: torch.device, 'cuda' or 'cpu'.

  Returns:
    optimizer: torch.optim, optimizer.
    model: model parameters.
    evaluation_loss: dictionary, contains loss components.
    evaluation_matching: list of namedtuple, contains metrics values.
  """
  n_sample1, p_feature1 = first_view.shape
  n_sample2, p_feature2 = second_view.shape

  cfg_model.sigmas = torch.FloatTensor([cfg_model.sigmas]).to(device)

  to_evaluate = (cfg_model.pca != 0
                 or cfg_model.n_record != 0 or cfg_model.n_eval != 0)

  def _train_and_evaluate(
      seed: int,
      evaluation: bool,
      inner_workdir: str = ''
      ) -> Any:

    evaluation_loss = ddict(list)
    evaluation_matching = ddict(list)
    pca_results = list()
    embeddings_results = list()

    torch.manual_seed(seed)
    random.seed(seed + 1)
    np.random.seed(seed + 2)

    summary_writer = None

    model = Model(cfg_model,
                  n_sample1,
                  n_sample2,
                  p_feature1,
                  p_feature2,
                  cfg_model.keops,
                  device)
    model = model.to(device)
    logging.info('seed=%s', seed)
    optimizer = torch.optim.Adam(model.parameters(), lr=cfg_model.learning_rate)
    model.train()

    # TODO(lpapaxanthos): stopping criterion.
    for i in range(cfg_model.n_iter):
      optimizer.zero_grad()
      loss, loss_components = model(first_view, second_view)
      loss.backward()
      optimizer.step()
      logging.info('Epoch: %s., train loss: %s.', i, loss)

      if evaluation:
        out = _evaluate(i, first_view, second_view, model, eval_fn, loss,
                        loss_components, evaluation_loss,
                        evaluation_matching, embeddings_results, pca_results,
                        cfg_model, device, summary_writer,
                        workdir=inner_workdir)
        evaluation_loss, evaluation_matching, embeddings_results, pca_results, summary_writer = out

    return (loss, optimizer, model, evaluation_loss, evaluation_matching,
            embeddings_results, pca_results)

  loss_output = np.inf
  seed_output = np.inf

  logging.info(cfg_model)

  # Sampling array of seeds.
  np.random.seed(cfg_model.seed)
  assert cfg_model.n_seed < 1e6, 'cfg_model.n_seed must be smaller than 1e6.'
  array_seeds = np.random.choice(int(1e6), cfg_model.n_seed, replace=False)

  if cfg_model.n_seed > 1:
    for k in array_seeds:
      loss, optimizer, model, *_ = _train_and_evaluate(
          k, evaluation=False, inner_workdir='')

      if loss < loss_output:
        loss_output = loss
        seed_output = k

  seed_output = seed_output if cfg_model.n_seed > 1 else cfg_model.seed
  out = _train_and_evaluate(seed_output,
                            evaluation=to_evaluate,
                            inner_workdir=workdir)
  _, optimizer, model, evaluation_loss, evaluation_matching, embeddings_results, pca_results = out
  return (
      optimizer,
      model,
      evaluation_loss,
      evaluation_matching,
      embeddings_results,
      pca_results,
      seed_output
      )



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates initializer function from flags."""
from torch import nn


def initialize(flag_init: str) -> nn.init:
  """Defines the initializer for the model parameters."""
  if flag_init == 'normal':
    initializer = nn.init.normal_
  elif flag_init == 'uniform':
    initializer = nn.init.uniform_
  elif flag_init == 'xavier_normal':
    initializer = nn.init.xavier_normal_
  elif flag_init == 'xavier_uniform':
    initializer = nn.init.xavier_uniform_
  elif not flag_init:
    initializer = nn.init.uniform_
  else:
    raise ValueError('Initializer not provided.')
  return initializer




# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Main module that generates the data and trains the model.

Instructions can be found on the readme.md document.
"""
import dataclasses
import os

from absl import app
from absl import flags
from absl import logging
from lsmmdma import train
from lsmmdma.data import checkpointer
from lsmmdma.data import data_pipeline
from lsmmdma.metrics import Evaluation
import numpy as np
import torch
import tensorflow as tf
from tensorflow.io import gfile

# Flags for input and output.
flags.DEFINE_string('output_dir', None, 'Output directory.')
flags.DEFINE_string('input_dir', None, 'Input directory.')
flags.DEFINE_string(
    'input_fv', None, 'Input first view, can be point cloud or kernel.')
flags.DEFINE_string(
    'input_sv', None, 'Input second view, can be point cloud or kernel.')
flags.DEFINE_string('rd_vec', None, 'Permutation of first view.')
flags.DEFINE_bool(
    'kernel', False, 'Whether the input is a point cloud or kernel.')
flags.DEFINE_enum('data', '', ['branch', 'triangle', ''],
                  'Chooses simulation or user input.')
flags.DEFINE_integer('n', 300, 'Sample size of generated data.')
flags.DEFINE_integer('p', 1000, 'Number of features in the generated data.')
flags.DEFINE_string('labels_fv', None,
                    'Filename for numerical cell labels for the first view.')
flags.DEFINE_string('labels_sv', None,
                    'Filename for numerical cell labels for the second view.')

# Random seeds.
flags.DEFINE_integer('seed', 0, 'Seed.')
flags.DEFINE_integer('ns', 1, 'Number of seeds with which to run the model.')

# Flags for training and evaluation.
flags.DEFINE_integer('e', 5001, 'Number of epochs.')
flags.DEFINE_integer('ne', 100, 'When to evaluate.')
flags.DEFINE_integer('nr', 100, 'When to record the loss.')
flags.DEFINE_integer('pca', 100, 'Applies PCA to embeddings every X epochs.')
flags.DEFINE_bool('short_eval', True,
                  'Whether or not to compute all the metrics.')
flags.DEFINE_integer('nn', 5,
                     'Number of neighbours in Label Transfer Accuracy metric.')

# Flags to define the algorithm.
flags.DEFINE_boolean('keops', True, 'Uses keops or not.')
flags.DEFINE_enum('m', 'dual', ['dual', 'primal'], 'Dual or primal mode.')
flags.DEFINE_bool('use_unbiased_mmd', True, 'Use unbiased MMD or not.')

# Flags for model hyperparameters.
flags.DEFINE_integer('d', 5, 'Dimension of output space.')
flags.DEFINE_string('init', 'uniform,0.,0.1', 'Initialiser.')
flags.DEFINE_float('l1', 1e-4, 'Hyperparameter for penalty terms.')
flags.DEFINE_float('l2', 1e-4, 'Hyperparameter for distortion terms.')
flags.DEFINE_float('lr', 1e-5, 'Learning rate.')
flags.DEFINE_float('s', 1.0, 'Scale parameter.')

# Flags to assess runtime
flags.DEFINE_bool('time', False, 'Whether or not to measure runtime.')


FLAGS = flags.FLAGS


def create_dummy_config(cfg: train.ModelGetterConfig):
  """Creates dummy config, needed when timing the training loop."""
  cfg = dataclasses.replace(cfg)
  cfg.n_iter = 1
  cfg.pca = 0
  cfg.n_record = 0
  cfg.n_eval = 0
  return cfg


def time_training_loop(func):
  """Enables to time the training loop."""
  def inner_fn(cfg_model: train.ModelGetterConfig, *args):
    cfg_model_time = create_dummy_config(cfg_model)
    _ = func(cfg_model_time, *args)
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start.record()
    out = func(cfg_model, *args)
    end.record()
    # Waits for everything to finish running
    torch.cuda.synchronize()
    runtime = start.elapsed_time(end)
    return runtime, out
  return inner_fn


def main(_):
  if FLAGS.kernel and FLAGS.m == 'primal':
    raise ValueError(f"""The flag kernel set to {FLAGS.kernel} is not compatible
                     with the flag mode set to {FLAGS.m}""")
  if ((FLAGS.labels_fv is not None and FLAGS.labels_sv is None)
      or (FLAGS.labels_fv is None and FLAGS.labels_sv is not None)):
    raise ValueError("""The flags labels_fv and labels_sv must be either both
                     defined or both None.""")

  tf.config.experimental.set_visible_devices([], 'GPU')
  logging.info('cuda is available: %s', torch.cuda.is_available())

  # Gets input data.
  if FLAGS.data:
    first_view, second_view, rd_vec = data_pipeline.generate_data(
        path=FLAGS.output_dir,
        random_seed=FLAGS.seed,
        n_sample=FLAGS.n,
        p_feature=FLAGS.p,
        simulation=FLAGS.data)
    labels = None
  else:
    first_view = data_pipeline.load(FLAGS.input_dir, FLAGS.input_fv)
    second_view = data_pipeline.load(FLAGS.input_dir, FLAGS.input_sv)
    if FLAGS.rd_vec:
      rd_vec = data_pipeline.load(FLAGS.input_dir, FLAGS.rd_vec)
    else:
      rd_vec = np.arange(first_view.shape[0])
    if FLAGS.labels_fv and FLAGS.labels_sv:
      labels_fv = data_pipeline.load(FLAGS.input_dir, FLAGS.labels_fv)
      labels_sv = data_pipeline.load(FLAGS.input_dir, FLAGS.labels_sv)
      labels = [labels_fv, labels_sv]
    else:
      labels = None

  # Creates output directory and filename.
  gfile.makedirs(FLAGS.output_dir)
  logging.info('output directory: %s', FLAGS.output_dir)
  filename = ':'.join(['m' + str(FLAGS.m),
                       'keops' + str(FLAGS.keops),
                       'ni' + str(FLAGS.e),
                       'seed' + str(FLAGS.seed),
                       'ns' + str(FLAGS.ns),
                       'lr' + str(FLAGS.lr),
                       's' + str(FLAGS.s),
                       'l1' + str(FLAGS.l1),
                       'l2' + str(FLAGS.l2),
                       'init' + str(FLAGS.init)])
  if FLAGS.data:
    filename = ':'.join([filename,
                         'n' + str(FLAGS.n),
                         'p' + str(FLAGS.p),
                         str(FLAGS.data)])
  else:
    filename = ':'.join([filename,
                         FLAGS.input_fv.split('.')[0]])

  # Chooses device.
  if torch.cuda.is_available():
    device = torch.device('cuda')
  else:
    device = 'cpu'

  # Prepares metrics.
  eval_fn = Evaluation(ground_truth_alignment=rd_vec,
                       cell_labels=labels,
                       device=device,
                       short=FLAGS.short_eval,
                       n_neighbours=FLAGS.nn)

  # Sets the hyperparameters of the model.
  cfg_model = train.ModelGetterConfig(
      seed=FLAGS.seed,
      n_seed=FLAGS.ns,
      low_dim=FLAGS.d,
      n_iter=FLAGS.e,
      keops=FLAGS.keops,
      mode=FLAGS.m,
      n_eval=FLAGS.ne,
      n_record=FLAGS.nr,
      learning_rate=FLAGS.lr,
      sigmas=FLAGS.s,
      lambda1=FLAGS.l1,
      lambda2=FLAGS.l2,
      pca=FLAGS.pca,
      init=FLAGS.init,
      use_unbiased_mmd=FLAGS.use_unbiased_mmd
      )

  # Prepares input.
  first_view = torch.FloatTensor(first_view).to(device)
  second_view = torch.FloatTensor(second_view).to(device)
  if cfg_model.mode == 'dual' and not FLAGS.kernel:
    first_view = train.get_kernel(first_view)
    second_view = train.get_kernel(second_view)

  # Runs model.
  train_fn = (time_training_loop(train.train_and_evaluate)
              if FLAGS.time else train.train_and_evaluate)
  if FLAGS.time:
    runtime, out = train_fn(
        cfg_model, first_view, second_view, eval_fn, '', device)
  else:
    out = train_fn(
        cfg_model, first_view, second_view, eval_fn, FLAGS.output_dir, device)
    runtime = '-'
  optim, model, eval_loss, eval_matching, emb_results, pca_results, seed = out

  # Saves results to files.
  loss = eval_loss['loss'][-1] if FLAGS.nr != 0 else -1
  mmd = eval_loss['mmd'][-1] if FLAGS.nr != 0 else -1
  foscttm = eval_matching['foscttm'][-1] if FLAGS.ne != 0 else -1
  top1 = eval_matching['top1'][-1] if FLAGS.ne != 0 else -1
  top5 = eval_matching['top5'][-1] if FLAGS.ne != 0 else -1
  no = eval_matching['no'][-1] if FLAGS.ne != 0 else -1
  lta = eval_matching['lta'][-1] if FLAGS.ne != 0 else -1
  results = [foscttm, top1, top5, no, lta]

  logging.info('Save results in %s.', FLAGS.output_dir)
  with gfile.GFile(
      os.path.join(FLAGS.output_dir, filename + '.tsv'), 'w') as my_file:
    colnames = ['model', 'seed', 'n_sample', 'n_feat', 'low_dim', 'n_iter',
                'keops', 'loss', 'mmd', 'foscttm', 'top1', 'top5',
                'no', 'lta', 'time']
    my_file.write('\t'.join(colnames) + '\n')
    checkpointer.save_data_eval(
        my_file, FLAGS, seed, loss, mmd, results, runtime, cfg_model)
  if FLAGS.ne != 0:
    logging.info('Save tracking in %s.', FLAGS.output_dir)
    checkpointer.save_tracking(FLAGS.output_dir,
                               filename,
                               eval_loss,
                               eval_matching,
                               seed,
                               FLAGS.e,
                               cfg_model)
  logging.info('Save model in %s.', FLAGS.output_dir)
  checkpointer.save_model(
      FLAGS.output_dir, filename, optim, model, seed, FLAGS.e, loss)
  logging.info('Save embeddings in %s', FLAGS.output_dir)
  checkpointer.save_embeddings(
      FLAGS.output_dir, filename, emb_results)
  if FLAGS.pca != 0:
    logging.info('Save PCA representation in %s', FLAGS.output_dir)
    checkpointer.save_pca(FLAGS.output_dir, filename, pca_results)
  logging.info('End.')

if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions used in MMDMA, PyTorch implementation."""
from typing import Optional, List
import numpy as np
import pykeops
pykeops.clean_pykeops()  # just in case old build files are still present
from pykeops.torch import LazyTensor
from sklearn import decomposition
import torch


# Functions in dual.
def pen_dual(
    embedding: torch.Tensor,
    param: torch.Tensor,
    device: torch.device
    ) -> torch.Tensor:
  """Computes the non-collapsing penalty for the dual formulation.

  In the dual formulation, when we project n points to low_dim dimensions,
  the model parameter 'param' is an n*low_dim matrix. The 'embedding'
  matrix is also an n*low_dim matrix equal to embedding=K*param,
  where K is the n*n kernel matrix of the n points in the input space.
  The 'embedding' matrix corresponds to the mapped set of points.
  The non-collapsing penalty is then equal to
  ||param.T * K * param - I ||_2, where I is the low_dim*low_dim
  identity matrix, which is also equal to ||param.T * embedding - I ||_2.

  Arguments:
    embedding: torch.Tensor, low dimensional representation of a view.
    param: torch.Tensor, model parameters.
    device: torch.device, whether cuda or cpu.

  Returns:
    penalty_value: torch.Tensor, scalar value.
  """
  low_dim = embedding.shape[1]
  identity_matrix_embedding_space = torch.eye(low_dim).to(device)
  penalty_value = (torch.matmul(param.t(), embedding)
                   - identity_matrix_embedding_space).norm(2)
  return penalty_value


def dis_dual(
    embedding: torch.Tensor,
    kernel: torch.Tensor
    ) -> torch.Tensor:
  """Computes the distortion penalty for the dual formulation.

  In the dual formulation, when we project n points to a shared space of
  low_dim dimensions, the 'embedding' matrix is an n*low_dim matrix equal to
  embedding=K*param, where K is the n*n kernel matrix of the n points in the
  input space and the n*low_dim 'param' matrix is the model parameter.  The
  'embedding' matrix corresponds to the mapped set of points.
  The distortion penalty is then equal to ||K * param * param.T * K - K ||_2,
  which is also equal to ||embedding * embedding.T - K||_2.

  Arguments:
    embedding: torch.Tensor, low dimensional representation of a view.
    kernel: torch.Tensor, view.

  Returns:
    penalty_value: torch.Tensor, scalar value.
  """
  distortion_value = (
      torch.matmul(embedding, embedding.t()) - kernel).norm(2)
  return distortion_value


# Functions for MMD.
def squared_mmd(
    first_view: torch.Tensor,
    second_view: torch.Tensor,
    sigmas: torch.Tensor,
    keops: bool,
    use_unbiased_mmd: bool,
    device: torch.device
    ) -> float:
  """Computes squared MMD with Gaussian kernel.

  Arguments:
    first_view: torch.Tensor, embedding for first view (sample x low_dim).
    second_view: torch.Tensor, embedding for second view (sample x low_dim).
    sigmas: torch.Tensor, scale parameter for Gaussian kernel.
    keops: bool, whether or not to use keops.
    use_unbiased_mmd: bool, whether to use (True) or not (False) the unbiased
      MMD.
    device: torch.device, device can be cuda or cpu.
  Returns:
    float, MMD value.
  """
  n_sample1 = first_view.shape[0]
  n_sample2 = second_view.shape[0]
  if use_unbiased_mmd:
    factor1 = 1.0 / (n_sample1 * (n_sample1 - 1.0))
    factor2 = 1.0 / (n_sample2 * (n_sample2 - 1.0))
    diagonal_term = 1.0 / (n_sample1 - 1.0) + 1.0 / (n_sample2 - 1.0)
  else:
    factor1 = 1.0 / n_sample1 ** 2
    factor2 = 1.0 / n_sample2 ** 2
    diagonal_term = 0

  if not keops:
    cost = gaussian_kernel(first_view, None, sigmas).sum() * factor1
    cost += gaussian_kernel(second_view, None, sigmas).sum() * factor2
    cost -= 2 * gaussian_kernel(first_view, second_view, sigmas).sum() / (
        n_sample1 * n_sample2)
    cost -= diagonal_term

  elif keops:
    # This section is largely inspired by https://github.com/jeanfeydy/geomloss.
    ones1 = torch.ones(n_sample1).to(device)
    ones2 = torch.ones(n_sample2).to(device)

    # (B,N,N) tensor
    kernel_11 = gaussian_kernel_keops(first_view, first_view, sigmas)
    # (B,M,M) tensor
    kernel_22 = gaussian_kernel_keops(second_view, second_view, sigmas)
    # (B,N,M) tensor
    kernel_12 = gaussian_kernel_keops(first_view, second_view, sigmas)

    # (B,N,N) @ (B,N) = (B,N)
    kernel_sum_1 = (kernel_11 @ ones1.unsqueeze(-1)).squeeze(-1)
    # (B,M,M) @ (B,M) = (B,M)
    kernel_sum_2 = (kernel_22 @ ones2.unsqueeze(-1)).squeeze(-1)
    # (B,N,M) @ (B,M) = (B,N)
    kernel_sum_12 = (kernel_12 @ ones2.unsqueeze(-1)).squeeze(-1)

    cost = (
        torch.dot(ones1.view(-1), kernel_sum_1.view(-1)) * factor1
        + torch.dot(ones2.view(-1), kernel_sum_2.view(-1)) * factor2
        - 2 * torch.dot(ones1.view(-1), kernel_sum_12.view(-1))
        / (n_sample1 * n_sample2) - diagonal_term
    )

  return cost if use_unbiased_mmd else torch.clip(cost, min=0)


def gaussian_kernel_keops(
    first_view: torch.Tensor,
    second_view: torch.Tensor,
    sigmas: torch.Tensor,
    ) -> float:
  """Computes Gaussian kernel.

  This function has been inspired from KeOps tutorial KeOps101
  (https://www.kernel-operations.io/keops/index.html).

  Arguments:
    first_view: torch.Tensor, embedding for first view (sample x low_dim).
    second_view: torch.Tensor, embedding for second view (sample x low_dim).
    sigmas: torch.Tensor, scale parameter of Gaussian kernel.

  Returns:
    Gaussian kernel.
  """
  beta = 1.0 / (2.0 * (sigmas.unsqueeze(1)))
  m, d = first_view.shape
  n, d = second_view.shape

  first_view = first_view * torch.sqrt(beta)
  second_view = second_view * torch.sqrt(beta)
  # Turn our dense Tensors into KeOps symbolic variables with "virtual"
  # dimensions at positions 0 and 1 (for "i" and "j" indices):
  first_view_i = LazyTensor(first_view.view(m, 1, d))
  second_view_j = LazyTensor(second_view.view(1, n, d))

  # We can now perform large-scale computations, without memory overflows:
  distance_ij = ((first_view_i - second_view_j)**2).sum(dim=2)
  # Symbolic (m, n, 1) matrix of squared distances
  kernel_ij = (- distance_ij).exp()
  return kernel_ij


def gaussian_kernel(
    first_view: torch.Tensor,
    second_view: Optional[torch.Tensor],
    sigmas: torch.Tensor,
    ) -> float:
  """Computes Gaussian kernel.

  Arguments:
    first_view: torch.Tensor, embedding for first view (sample x low_dim).
    second_view: torch.Tensor, embedding for second view (sample x low_dim).
      When it is set to None, the kernel is calculated between the first_view
      and itself.
    sigmas: torch.Tensor, scale parameter of Gaussian kernel.

  Returns:
    Gaussian kernel.
  """
  beta = 1. / (2. * sigmas)
  distances = compute_sqpairwise_distances(first_view, second_view)
  kernel = torch.exp(- beta * distances)
  return kernel


# Functions in primal.
def pen_primal(param: torch.Tensor, device: torch.device) -> torch.Tensor:
  """Computes non-collapsing penalty for the primal formulation.

  In the primal formulation, the 'param' matrix is the n*low_dim model parameter
  and the non-collapsing penalty is equal to ||param.T * param - I||_2,
  where I is the low_dim*low_dim identity matrix.

  Arguments:
    param: torch.Tensor, model parameters.
    device: torch.device, whether cuda or cpu.
  Returns:
    penalty_value: torch.Tensor, scalar value.
  """
  low_dim = param.shape[1]
  identity_matrix_embedding_space = torch.eye(low_dim).to(device)
  penalty_value = (torch.matmul(param.t(), param)
                   - identity_matrix_embedding_space).norm(2)
  return penalty_value


def dis_primal(
    input_view: torch.Tensor,
    param: torch.Tensor,
    n_sample: int,
) -> torch.Tensor:
  """Computes distortion penalty for the primal formulation.

  Let n be the number of samples in the view of interest and p the number of
  features. In the primal formulation, the 'param' matrix is the p*low_dim model
  parameter and input_view corresponds to the input data, of shape n*p. The
  distortion penalty can be written as

  distortion = ||input_view*input_view.T
                 - input_view*param*param.T*input_view.T||_2.

  The distortion is computed as is when n < p. However, if n > p, we
  compute the following formulation:

  distortion = torch.sqrt(Tr((I - param*param.T)*input_view.T*input_view
  *(I - param*param.T)*input_view.T*input_view))

  to avoid computing terms that are O(n**2) in memory or runtime.

  Arguments:
    input_view: torch.Tensor, one of the two views.
    param: torch.Tensor, model parameters.
    n_sample: int, sample size of entire dataset.
  Returns:
    distortion_value: torch.Tensor, scalar value.
  """
  n_sample, p_feature = input_view.shape
  if n_sample < p_feature:
    inner_prod = torch.matmul(input_view, input_view.t())
    tmp = torch.matmul(torch.matmul(
        torch.matmul(input_view, param), param.t()), input_view.t())
    tmp = (inner_prod - tmp)**2
    distortion_value = torch.sqrt(torch.sum(tmp))
  else:
    gram = torch.matmul(input_view.t(), input_view)
    tmp = torch.matmul(param, torch.matmul(param.t(), gram))
    prod = gram - tmp
    distortion_value = torch.sqrt(torch.trace(torch.matmul(prod, prod)))
  return distortion_value


# Others.
def compute_sqpairwise_distances(
    first_view: torch.Tensor,
    second_view: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
  """Computes squared euclidean pairwise distances.

  Args:
    first_view: numpy array, format is samples_x x features.
    second_view: numpy array, format is samples_y x features.
  Returns:
    jax numpy array, format is samples_y x samples_x.
  """
  if second_view is None:
    sample_norm_first_view = torch.sum(
        torch.square(first_view), axis=1).reshape(-1, 1)
    cross_inner_product = torch.matmul(first_view, first_view.t())
    diff = (sample_norm_first_view
            + sample_norm_first_view.reshape(1, -1) - 2 * cross_inner_product)
    diff -= torch.diag(torch.diag(diff))
  else:
    sample_norm_first_view = torch.sum(
        torch.square(first_view), axis=1).reshape(-1, 1)
    sample_norm_second_view = torch.sum(
        torch.square(second_view), axis=1).reshape(1, -1)
    cross_inner_product = torch.matmul(first_view, second_view.t())
    diff = (sample_norm_first_view
            + sample_norm_second_view - 2 * cross_inner_product)
  diff = torch.clip(diff, min=0)
  return diff


# Output function.
def pca(first_view, second_view, low_dim=2):
  """Project the samples into a lower dimension with PCA."""
  pca_fn = decomposition.PCA(n_components=low_dim)
  n = first_view.shape[0]
  views = np.concatenate((first_view, second_view), axis=0)
  pca_rep = pca_fn.fit_transform(views)
  pca_rep_fv = pca_rep[:n]
  pca_rep_sv = pca_rep[n:]
  return pca_rep_fv, pca_rep_sv





# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script for installing large_scale_mmdma as a pip module."""
import os
from setuptools import find_packages
from setuptools import setup

folder = os.path.dirname(__file__)
# Reads the version
__version__ = None
with open(os.path.join(folder, "lsmmdma/version.py")) as f:
  exec(f.read(), globals())

readme_path = os.path.join(folder, "README.md")
readme_contents = ""
if os.path.exists(readme_path):
  with open(readme_path) as fp:
    readme_contents = fp.read().strip()

# Reads the requirements from requirements.txt
folder = os.path.dirname(__file__)
path = os.path.join(folder, "requirements.txt")
install_requires = []
if os.path.exists(path):
  with open(path) as fp:
    install_requires = [line.strip() for line in fp]


setup(
    name="lsmmdma",
    version=__version__,
    description="Scaling MMD-MA.",
    author="Google LLC",
    author_email="no-reply@google.com",
    url="https://github.com/google-research/large_scale_mmdma",
    long_description=readme_contents,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    packages=find_packages(),
    package_data={},
    install_requires=install_requires,
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
    ],
    keywords="kernel, data integration",
    requires_python=">=3.6",
)





# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Saves results."""
import dataclasses
import json
import os
from typing import Tuple, DefaultDict, List

import absl
from lsmmdma.train import ModelGetterConfig
import numpy as np
import torch
from tensorflow.io import gfile


def save_generated_data(
    path: str,
    first_view: torch.FloatTensor,
    second_view: torch.FloatTensor,
    rd_vec: np.ndarray):
  """Saves generated data."""
  file_prefix = 'generated_data'
  gfile.makedirs(path)
  with gfile.GFile(os.path.join(path, file_prefix + '_fv.pt'), 'w') as my_file:
    torch.save(first_view, my_file)
  with gfile.GFile(os.path.join(path, file_prefix + '_sv.pt'), 'w') as my_file:
    torch.save(second_view, my_file)
  with gfile.GFile(
      os.path.join(path, file_prefix + '_rd_vec.npy'), 'wb') as my_file:
    np.save(my_file, rd_vec)


def save_data_eval(
    my_file: gfile.GFile,
    flags: absl.flags,
    seed: int,
    loss: float,
    mmd: float,
    res: float,
    runtime: float,
    cfg_model: ModelGetterConfig,
    ):
  """Saves main results."""
  cfg_dict = dataclasses.asdict(cfg_model)
  val = cfg_dict.get('keops', None)
  args = [flags.m, seed, flags.n, flags.p, flags.d,
          flags.e, val, loss, mmd, *res, runtime]
  my_file.write('\t'.join(map(str, args)) + '\n')


def save_model(
    path: str,
    filename: str,
    optimizer: torch.optim,
    model: torch.nn.Module,
    seed: int,
    epoch: int,
    loss: float
    ):
  """Saves model parameters."""
  filename = f'{filename}_model.json'
  with gfile.GFile(os.path.join(path, filename), 'w') as my_file:
    torch.save({
        'seed': seed,
        'epoch': epoch,
        'model_state_dict': model.state_dict(),
        'optimizer_state_dict': optimizer.state_dict(),
        'loss': loss
        }, my_file)


def save_pca(
    path: str,
    filename: str,
    pca_results: List[np.ndarray],
    ):
  """Saves 2D representation."""
  filename_pca = f'{filename}_pca'
  pca_fv = [element[0] for element in pca_results]
  pca_sv = [element[1] for element in pca_results]
  with gfile.GFile(os.path.join(path, filename_pca + '_fv.npy'), 'w') as myfile:
    np.save(myfile, np.array(pca_fv))
  with gfile.GFile(os.path.join(path, filename_pca + '_sv.npy'), 'w') as myfile:
    np.save(myfile, np.array(pca_sv))


def save_embeddings(
    path: str,
    filename: str,
    embeddings_results: List[np.ndarray],
    ):
  """Saves low dimensional representation."""
  filename_emb = f'{filename}_embeddings'
  embeddings_fv = [element[0] for element in embeddings_results]
  embeddings_sv = [element[1] for element in embeddings_results]
  with gfile.GFile(os.path.join(path, filename_emb + '_fv.npy'), 'w') as myfile:
    np.save(myfile, np.array(embeddings_fv))
  with gfile.GFile(os.path.join(path, filename_emb + '_sv.npy'), 'w') as myfile:
    np.save(myfile, np.array(embeddings_sv))


def save_tracking(
    path: str,
    filename: str,
    evaluation_loss: DefaultDict[str, List[float]],
    evaluation_matching: DefaultDict[str, List[float]],
    seed: int,
    epoch: int,
    cfg_model: ModelGetterConfig
    ):
  """Saves evaluation measures."""
  filename = f'{filename}_tracking.json'
  with gfile.GFile(os.path.join(path, filename), 'w') as my_file:
    data = {
        'seed': int(seed),
        'mode': cfg_model.mode,
        'keops': cfg_model.keops,
        'epoch': epoch,
        'evaluation_loss': evaluation_loss,
        'evaluation_matching': evaluation_matching
    }
    json.dump(data, my_file)
